# ruff: noqa: SLF001
from pathlib import Path

import pytest

from darkseid.archivers import UnknownArchiver
from darkseid.archivers.rar import RarArchiver
from darkseid.archivers.zip import ZipArchiver
from darkseid.comic import Comic
from darkseid.metadata import Metadata


@pytest.mark.parametrize(
    ("path", "expected_archiver"),
    [
        ("/path/to/comic.cbz", ZipArchiver),
        ("/path/to/comic.cbr", RarArchiver),
        ("/path/to/comic.unknown", UnknownArchiver),
    ],
    ids=["zip file", "rar file", "unknown file"],
)
def test_comic_initialization(mocker, path, expected_archiver):
    # Arrange
    mocker.patch("zipfile.is_zipfile", return_value=path.endswith(".cbz"))
    mocker.patch("rarfile.is_rarfile", return_value=path.endswith(".cbr"))

    # Act
    comic = Comic(path)

    # Assert
    assert isinstance(comic.archiver, expected_archiver)


def test_comic_str():
    # Arrange
    path = "/path/to/comic.cbz"
    comic = Comic(path)

    # Act
    result = str(comic)

    # Assert
    assert result == "comic.cbz"


def test_comic_path():
    # Arrange
    path = "/path/to/comic.cbz"
    comic = Comic(path)

    # Act
    result = comic.path

    # Assert
    assert result == Path(path)


def test_reset_cache():
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    comic._has_md = True
    comic._page_count = 10
    comic._page_list = ["page1", "page2"]
    comic._metadata = Metadata()

    # Act
    comic.reset_cache()

    # Assert
    assert comic._has_md is None
    assert comic._page_count is None
    assert comic._page_list is None
    assert comic._metadata is None


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        ("/path/to/comic.cbr", True),
        ("/path/to/comic.cbz", False),
    ],
    ids=["rar file", "not rar file"],
)
def test_rar_test(mocker, path, expected):
    # Arrange
    mocker.patch("rarfile.is_rarfile", return_value=path.endswith(".cbr"))
    comic = Comic(path)

    # Act
    result = comic.rar_test()

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("path", "expected"),
    [
        ("/path/to/comic.cbz", True),
        ("/path/to/comic.cbr", False),
    ],
    ids=["zip file", "not zip file"],
)
def test_zip_test(mocker, path, expected):
    # Arrange
    mocker.patch("zipfile.is_zipfile", return_value=path.endswith(".cbz"))
    comic = Comic(path)

    # Act
    result = comic.zip_test()

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("archive_type", "expected"),
    [
        (Comic.ArchiveType.rar, True),
        (Comic.ArchiveType.zip, False),
    ],
    ids=["rar archive", "not rar archive"],
)
def test_is_rar(archive_type, expected):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    comic._archive_type = archive_type

    # Act
    result = comic.is_rar()

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("archive_type", "expected"),
    [
        (Comic.ArchiveType.zip, True),
        (Comic.ArchiveType.rar, False),
    ],
    ids=["zip archive", "not zip archive"],
)
def test_is_zip(archive_type, expected):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    comic._archive_type = archive_type

    # Act
    result = comic.is_zip()

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("archive_type", "expected"),
    [
        (Comic.ArchiveType.zip, True),
        (Comic.ArchiveType.rar, False),
        (Comic.ArchiveType.unknown, False),
    ],
    ids=["zip writable", "rar not writable", "unknown not writable"],
)
def test_is_writable(mocker, archive_type, expected):
    # Arrange
    path = "/path/to/comic.cbz"
    comic = Comic(path)
    comic._archive_type = archive_type
    mocker.patch("os.access", return_value=True)

    # Act
    result = comic.is_writable()

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("is_zip", "is_rar", "page_count", "expected"),
    [
        (True, False, 5, True),
        (False, True, 5, True),
        (False, False, 5, False),
        (True, False, 0, False),
    ],
    ids=["zip with pages", "rar with pages", "not zip or rar", "zip with no pages"],
)
def test_seems_to_be_a_comic_archive(mocker, is_zip, is_rar, page_count, expected):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "is_zip", return_value=is_zip)
    mocker.patch.object(comic, "is_rar", return_value=is_rar)
    mocker.patch.object(comic, "get_number_of_pages", return_value=page_count)

    # Act
    result = comic.seems_to_be_a_comic_archive()

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("index", "expected"),
    [
        (0, b"image data"),
        (1, None),
    ],
    ids=["valid index", "invalid index"],
)
def test_get_page(mocker, index, expected):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "get_page_name", return_value="page1.jpg" if index == 0 else None)
    mocker.patch.object(comic._archiver, "read_file", return_value=b"image data")

    # Act
    result = comic.get_page(index)

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("index", "expected"),
    [
        (0, "page1.jpg"),
        (1, None),
    ],
    ids=["valid index", "invalid index"],
)
def test_get_page_name(mocker, index, expected):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "get_page_name_list", return_value=["page1.jpg"])

    # Act
    result = comic.get_page_name(index)

    # Assert
    assert result == expected


@pytest.mark.parametrize(
    ("name_path", "expected"),
    [
        (Path("image.jpg"), True),
        (Path("image.txt"), False),
        (Path(".hidden.jpg"), False),
    ],
    ids=["valid image", "invalid image", "hidden image"],
)
def test_is_image(name_path, expected):
    # Act
    result = Comic.is_image(name_path)

    # Assert
    assert result == expected


def test_get_page_name_list(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(
        comic._archiver,
        "get_filename_list",
        return_value=["page1.jpg", "page2.png", "not_image.txt"],
    )
    mocker.patch(
        "darkseid.comic.Comic.is_image", side_effect=lambda x: x.suffix in [".jpg", ".png"]
    )

    # Act
    result = comic.get_page_name_list()

    # Assert
    assert result == ["page1.jpg", "page2.png"]


def test_get_number_of_pages(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "get_page_name_list", return_value=["page1.jpg", "page2.png"])

    # Act
    result = comic.get_number_of_pages()

    # Assert
    assert result == 2


def test_read_metadata(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "read_raw_metadata", return_value="<ComicInfo></ComicInfo>")
    mocker.patch("darkseid.comic.ComicInfo.metadata_from_string", return_value=Metadata())

    # Act
    result = comic.read_metadata()

    # Assert
    assert isinstance(result, Metadata)


def test_read_raw_metadata(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "has_metadata", return_value=True)
    mocker.patch.object(comic._archiver, "read_file", return_value=b"<ComicInfo></ComicInfo>")

    # Act
    result = comic.read_raw_metadata()

    # Assert
    assert result == "<ComicInfo></ComicInfo>"


def test_write_metadata(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    metadata = Metadata()
    mocker.patch.object(comic, "is_writable", return_value=True)
    mocker.patch.object(comic, "apply_archive_info_to_metadata")
    mocker.patch.object(comic, "read_raw_metadata", return_value=None)
    mocker.patch(
        "darkseid.comic.ComicInfo.string_from_metadata", return_value="<ComicInfo></ComicInfo>"
    )
    mocker.patch.object(comic._archiver, "write_file", return_value=True)
    mocker.patch.object(comic, "_successful_write", return_value=True)

    # Act
    result = comic.write_metadata(metadata)

    # Assert
    assert result is True


def test_remove_metadata(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "has_metadata", return_value=True)
    mocker.patch.object(comic._archiver, "remove_file", return_value=True)
    mocker.patch.object(comic, "_successful_write", return_value=True)

    # Act
    result = comic.remove_metadata()

    # Assert
    assert result is True


def test_remove_pages(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "get_page_name", side_effect=["page1.jpg", "page2.png"])
    mocker.patch.object(comic._archiver, "remove_files", return_value=True)
    mocker.patch.object(comic, "_successful_write", return_value=True)

    # Act
    result = comic.remove_pages([0, 1])

    # Assert
    assert result is True


def test_has_metadata(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    mocker.patch.object(comic, "seems_to_be_a_comic_archive", return_value=True)
    mocker.patch.object(comic._archiver, "get_filename_list", return_value=["ComicInfo.xml"])

    # Act
    result = comic.has_metadata()

    # Assert
    assert result is True


def test_apply_archive_info_to_metadata(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbz")
    metadata = Metadata()
    mocker.patch.object(comic, "get_number_of_pages", return_value=2)
    mocker.patch.object(comic, "get_page", return_value=b"image data")
    mocker.patch("PIL.Image.open", return_value=mocker.Mock(size=(100, 200)))

    # Act
    comic.apply_archive_info_to_metadata(metadata, calc_page_sizes=True)

    # Assert
    assert metadata.page_count == 2


def test_export_as_zip(mocker):
    # Arrange
    comic = Comic("/path/to/comic.cbr")
    mocker.patch("darkseid.archivers.zip.ZipArchiver.copy_from_archive", return_value=True)

    # Act
    result = comic.export_as_zip(Path("/path/to/comic.cbz"))

    # Assert
    assert result is True



from datetime import date
from decimal import Decimal

import pytest

from darkseid.metadata import Arc, Basic, Metadata, Price, Series, Universe


@pytest.fixture(scope="module")
def fake_metadata() -> Metadata:
    md = Metadata()
    md.series = Series(
        name="Aquaman",
        sort_name="Aquaman",
        volume=1,
        format="Annual",
    )
    md.issue = "0"
    md.stories = [Basic("A Crash of Symbols")]
    md.publisher = Basic("DC Comics")
    md.imprint = Basic("Vertigo", 9)
    md.cover_date = date(1994, 12, 1)
    md.story_arcs = [Arc("Final Crisis, Inc")]
    md.characters = [
        Basic("Aquaman"),
        Basic("Mera"),
        Basic("Garth"),
    ]
    md.teams = [Basic("Justice League"), Basic("Infinity, Inc")]
    md.universes = [Universe(id_=25, name="ABC", designation="Earth 25")]
    md.comments = "Just some sample metadata."
    md.black_and_white = True
    md.is_empty = False

    return md


@pytest.fixture(scope="session")
def fake_overlay_metadata() -> Metadata:
    overlay_md = Metadata()
    overlay_md.series = Series(name="Aquaman", sort_name="Aquaman", volume=1, format="Annual")
    overlay_md.cover_date = date(1994, 10, 1)
    overlay_md.reprints = [Basic("Aquaman (1964) #64", 12345)]
    overlay_md.prices = [Price(Decimal("3.99")), Price(Decimal("1.5"), "CA")]
    overlay_md.collection_title = "Just another TPB"
    return overlay_md



# ruff: noqa: C901, PLR0912, PLR0915, TRY003, EM101
"""A class to encapsulate ComicRack's ComicInfo.xml data."""

# Copyright 2012-2014 Anthony Beville
# Copyright 2020 Brian Pepple
from __future__ import annotations

import re
import xml.etree.ElementTree as ET  # noqa: N817
from datetime import date
from typing import Any, ClassVar, cast

from defusedxml.ElementTree import fromstring, parse

from darkseid.issue_string import IssueString
from darkseid.metadata import Arc, Basic, Credit, ImageMetadata, Metadata, Role, Series
from darkseid.utils import list_to_string, xlate


class ComicInfo:
    """
    Handles the conversion between Metadata objects and XML representations.

    Includes methods for converting Metadata to XML, XML to Metadata, and writing/reading Metadata to/from external files.
    """

    ci_age_ratings: ClassVar[frozenset[str]] = frozenset(
        {
            "Unknown",
            "Adults Only 18+",
            "Early Childhood",
            "Everyone",
            "Everyone 10+",
            "G",
            "Kids to Adults",
            "M",
            "MA15+",
            "Mature 17+",
            "PG",
            "R18+",
            "Rating Pending",
            "Teen",
            "X18+",
        }
    )

    ci_manga: ClassVar[frozenset[str]] = frozenset({"Unknown", "Yes", "No", "YesAndRightToLeft"})

    writer_synonyms: ClassVar[frozenset[str]] = frozenset(
        {
            "writer",
            "plotter",
            "scripter",
            "script",
            "story",
            "plot",
        }
    )
    penciller_synonyms: ClassVar[frozenset[str]] = frozenset(
        {
            "artist",
            "breakdowns",
            "illustrator",
            "layouts",
            "penciller",
            "penciler",
        }
    )
    inker_synonyms: ClassVar[frozenset[str]] = frozenset(
        {
            "artist",
            "embellisher",
            "finishes",
            "illustrator",
            "ink assists",
            "inker",
        }
    )
    colorist_synonyms: ClassVar[frozenset[str]] = frozenset(
        {
            "colorist",
            "colourist",
            "colorer",
            "colourer",
            "color assists",
            "color flats",
        }
    )
    letterer_synonyms: ClassVar[frozenset[str]] = frozenset({"letterer"})
    cover_synonyms: ClassVar[frozenset[str]] = frozenset(
        {
            "cover",
            "covers",
            "coverartist",
            "cover artist",
        }
    )
    editor_synonyms: ClassVar[frozenset[str]] = frozenset(
        {
            "assistant editor",
            "associate editor",
            "consulting editor",
            "editor",
            "editor in chief",
            "executive editor",
            "group editor",
            "senior editor",
            "supervising editor",
        }
    )

    def metadata_from_string(self: ComicInfo, string: str) -> Metadata:
        """
        Parses an XML string representation into a Metadata object.

        Args:
            string (str): The XML string to parse.

        Returns:
            Metadata: The parsed Metadata object.
        """
        tree = ET.ElementTree(fromstring(string))
        return self.convert_xml_to_metadata(tree)

    def string_from_metadata(
        self: ComicInfo,
        md: Metadata,
        xml: bytes = b"",
    ) -> str:
        """
        Converts Metadata object to an XML string representation.

        Args:
            md (Metadata): The Metadata object to convert.
            xml (bytes): Additional XML content, defaults to an empty byte string.

        Returns:
            str: The XML string representation of the Metadata object.
        """
        tree = self.convert_metadata_to_xml(md, xml)
        return ET.tostring(tree.getroot(), encoding="utf-8", xml_declaration=True).decode()

    @classmethod
    def _split_sting(cls: type[ComicInfo], string: str, delimiters: list[str]) -> list[str]:
        """
        Splits a string based on the provided delimiters.

        Args:
            string (str): The string to split.
            delimiters (list[str]): List of delimiters to use for splitting.

        Returns:
            list[str]: The list of substrings after splitting the string.
        """
        for delimiter in delimiters:
            string = string.replace(delimiter, delimiters[0])
        return string.split(delimiters[0])

    @staticmethod
    def _get_root(xml: any) -> ET.Element:
        """
        Returns the root element of an XML object.

        Args:
            xml (any): The XML object to extract the root element from.

        Returns:
            ET.Element: The root element of the XML object.
        """
        root = ET.ElementTree(fromstring(xml)).getroot() if xml else ET.Element("ComicInfo")
        root.attrib["xmlns:xsi"] = "https://www.w3.org/2001/XMLSchema-instance"
        root.attrib["xmlns:xsd"] = "https://www.w3.org/2001/XMLSchema"

        return root

    @classmethod
    def validate_value(
        cls: type[ComicInfo], val: str | None, valid_set: frozenset[str]
    ) -> str | None:
        """
        Validates a value against a predefined set.

        Args:
            val (str | None): The value to validate.
            valid_set (frozenset[str]): The set of valid values.

        Returns:
            str | None: The validated value, or "Unknown" if the value is not in the set.
        """
        if val is not None:
            return "Unknown" if val not in valid_set else val
        return None

    def convert_metadata_to_xml(
        self: ComicInfo,
        md: Metadata,
        xml: bytes = b"",
    ) -> ET.ElementTree:
        """
        Converts Metadata object to an XML representation.

        Args:
            md (Metadata): The Metadata object to convert.
            xml (bytes): Additional XML content, defaults to an empty byte string.

        Returns:
            ET.ElementTree: The XML representation of the Metadata object.
        """
        root = self._get_root(xml)

        def assign(cix_entry: str, md_entry: str | int | None) -> None:
            et_entry = root.find(cix_entry)
            if md_entry is not None and md_entry:
                if et_entry is not None:
                    et_entry.text = str(md_entry)
                else:
                    ET.SubElement(root, cix_entry).text = str(md_entry)
            elif et_entry is not None:
                root.remove(et_entry)

        def get_resource_list(resource: list[Basic] | list[Arc]) -> str | None:
            return list_to_string([i.name for i in resource]) if resource else None

        assign("Title", get_resource_list(md.stories))
        if md.series is not None:
            assign("Series", md.series.name)
        assign("Number", md.issue)
        assign("Count", md.issue_count)
        if md.series is not None:
            assign("Volume", md.series.volume)
        assign("AlternateSeries", md.alternate_series)
        assign("AlternateNumber", md.alternate_number)
        assign("SeriesGroup", md.series_group)
        assign("AlternateCount", md.alternate_count)
        assign("Summary", md.comments)
        assign("Notes", md.notes)
        if md.cover_date is not None:
            assign("Year", md.cover_date.year)
            assign("Month", md.cover_date.month)
            assign("Day", md.cover_date.day)

        credit_roles = {
            "Writer": self.writer_synonyms,
            "Penciller": self.penciller_synonyms,
            "Inker": self.inker_synonyms,
            "Colorist": self.colorist_synonyms,
            "Letterer": self.letterer_synonyms,
            "CoverArtist": self.cover_synonyms,
            "Editor": self.editor_synonyms,
        }

        credit_lists = {role: [] for role in credit_roles}

        for credit in md.credits:
            for r in credit.role:
                role_name = r.name.casefold()
                for role, synonyms in credit_roles.items():
                    if role_name in synonyms:
                        credit_lists[role].append(credit.person.replace(",", ""))

        for role, names in credit_lists.items():
            assign(role, list_to_string(names))

        if md.publisher:
            assign("Publisher", md.publisher.name)
        if md.imprint:
            assign("Imprint", md.imprint.name)
        assign("Genre", get_resource_list(md.genres))
        assign("Web", md.web_link)
        assign("PageCount", md.page_count)
        if md.series is not None:
            assign("LanguageISO", md.series.language)
        if md.series is not None:
            assign("Format", md.series.format)
        assign("BlackAndWhite", "Yes" if md.black_and_white else None)
        assign("Manga", self.validate_value(md.manga, self.ci_manga))
        assign("Characters", get_resource_list(md.characters))
        assign("Teams", get_resource_list(md.teams))
        assign("Locations", get_resource_list(md.locations))
        assign("ScanInformation", md.scan_info)
        assign("StoryArc", get_resource_list(md.story_arcs))
        assign("AgeRating", self.validate_value(md.age_rating, self.ci_age_ratings))

        #  loop and add the page entries under pages node
        pages_node = root.find("Pages")
        if pages_node is not None:
            pages_node.clear()
        else:
            pages_node = ET.SubElement(root, "Pages")

        for page_dict in md.pages:
            page_dict["Image"] = str(page_dict.get("Image", ""))
            page_node = ET.SubElement(pages_node, "Page")
            page_node.attrib = dict(sorted(page_dict.items()))

        ET.indent(root)
        return ET.ElementTree(root)

    def convert_xml_to_metadata(self: ComicInfo, tree: ET.ElementTree) -> Metadata:
        """
        Converts an XML representation to a Metadata object.

        Args:
            tree (ET.ElementTree): The XML tree to convert to Metadata.

        Returns:
            Metadata: The Metadata object extracted from the XML tree.
        """
        root = tree.getroot()

        if root.tag != "ComicInfo":
            raise ValueError("Metadata is not ComicInfo format")

        def get(txt: str) -> str | int | None:
            """
            Finds and returns the text content of a specific tag in an XML tree.

            Args:
                txt (str): The tag to search for.

            Returns:
                str | int | None: The text content of the tag, or None if the tag is not found.
            """
            tag = root.find(txt)
            return None if tag is None else tag.text

        md = Metadata()
        md.series = Series(name=xlate(get("Series")))
        md.stories = self.string_to_resource(xlate(get("Title")))
        md.issue = IssueString(xlate(get("Number"))).as_string()
        md.issue_count = xlate(get("Count"), True)
        md.series.volume = xlate(get("Volume"), True)
        md.alternate_series = xlate(get("AlternateSeries"))
        md.alternate_number = IssueString(xlate(get("AlternateNumber"))).as_string()
        md.alternate_count = xlate(get("AlternateCount"), True)
        md.comments = xlate(get("Summary"))
        md.notes = xlate(get("Notes"))
        # Cover Year
        tmp_year = xlate(get("Year"), True)
        tmp_month = xlate(get("Month"), True)
        tmp_day = xlate(get("Day"), True)
        if tmp_year is not None and tmp_month is not None:
            if tmp_day is not None:
                md.cover_date = date(tmp_year, tmp_month, tmp_day)
            else:
                md.cover_date = date(tmp_year, tmp_month, 1)

        md.publisher = Basic(xlate(get("Publisher")))
        if imprint := xlate(get("Imprint")):  # Make sure Imprint element is present.
            md.imprint = Basic(imprint)
        md.genres = self.string_to_resource(xlate(get("Genre")))
        md.web_link = xlate(get("Web"))
        md.series.language = xlate(get("LanguageISO"))
        md.series.format = xlate(get("Format"))
        md.manga = xlate(get("Manga"))
        md.characters = self.string_to_resource(xlate(get("Characters")))
        md.teams = self.string_to_resource(xlate(get("Teams")))
        md.locations = self.string_to_resource(xlate(get("Locations")))
        md.page_count = xlate(get("PageCount"), True)
        md.scan_info = xlate(get("ScanInformation"))
        md.story_arcs = self.string_to_arc(xlate(get("StoryArc")))
        md.series_group = xlate(get("SeriesGroup"))
        md.age_rating = xlate(get("AgeRating"))

        tmp = xlate(get("BlackAndWhite"))
        md.black_and_white = False
        if tmp is not None and tmp.casefold() in ["yes", "true", "1"]:
            md.black_and_white = True
        # Now extract the credit info
        for n in root:
            if (
                n.tag in ["Writer", "Penciller", "Inker", "Colorist", "Letterer", "Editor"]
                and n.text is not None
            ):
                for name in self._split_sting(n.text, [";"]):
                    md.add_credit(Credit(name.strip(), [Role(n.tag)]))

            if n.tag == "CoverArtist" and n.text is not None:
                for name in self._split_sting(n.text, [";"]):
                    md.add_credit(Credit(name.strip(), [Role("Cover")]))

        # parse page data now
        pages_node = root.find("Pages")
        if pages_node is not None:
            for page in pages_node:
                p: dict[str, Any] = page.attrib
                if "Image" in p:
                    p["Image"] = int(p["Image"])
                md.pages.append(cast(ImageMetadata, p))

        md.is_empty = False

        return md

    def write_to_external_file(
        self: ComicInfo,
        filename: str,
        md: Metadata,
        xml: bytes = b"",
    ) -> None:
        """
        Writes Metadata to an external file in XML format.

        Args:
            filename (str): The name of the file to write the Metadata to.
            md (Metadata): The Metadata object to write to the file.
            xml (bytes): Additional XML content, defaults to an empty byte string.

        Returns:
            None
        """
        tree = self.convert_metadata_to_xml(md, xml)
        tree.write(filename, encoding="utf-8", xml_declaration=True)

    def read_from_external_file(self: ComicInfo, filename: str) -> Metadata:
        """
        Reads Metadata from an external file in XML format.

        Args:
            filename (str): The name of the file to read the Metadata from.

        Returns:
            Metadata: The Metadata object extracted from the file.
        """
        tree = parse(filename)
        return self.convert_xml_to_metadata(tree)

    @staticmethod
    def clean_resource_list(string: str) -> list[str]:
        """
        Cleans and filters a string to create a list of non-empty values.

        Args:
            string (str): The string to clean and filter.

        Returns:
            list[str]: The list of cleaned and filtered non-empty values.
        """
        return [item.strip() for item in re.split(r',|"(.*?)"', string) if item and item.strip()]

    @staticmethod
    def string_to_resource(string: str) -> list[Basic] | None:
        """
        Converts a string to a list of Basic objects.

        Args:
            string (str): The string to convert to Basic objects.

        Returns:
            list[Basic] | None: The list of Basic objects created from the string, or None if the string is None.
        """
        if string is not None:
            res: list[str | Basic] = ComicInfo.clean_resource_list(string)
            return [Basic(item) for item in res]
        return None

    @staticmethod
    def string_to_arc(string: str) -> list[Arc] | None:
        """
        Converts a string to a list of Arc objects.

        Args:
            string (str): The string to convert to Arc objects.

        Returns:
            list[Arc] | None: The list of Arc objects created from the string, or None if the string is None.
        """
        if string is not None:
            res: list[str | Arc] = ComicInfo.clean_resource_list(string)
            return [Arc(item) for item in res]
        return None



from datetime import date
from decimal import Decimal

import pytest

from darkseid.metadata import GTIN, Credit, Metadata, Price, Role, Series

MARTY = "Martin Egeland"
PETER = "Peter David"
WRITER = "Writer"
PENCILLER = "Penciller"
COVER = "Cover"


def test_metadata_print_str(fake_metadata: Metadata) -> None:
    expect_res = """Metadata(
    is_empty = False,
    series = Series(name='Aquaman', id_=None, sort_name='Aquaman', volume=1, format='Annual', language=None),
    issue = '0',
    stories = [Basic(name='A Crash of Symbols', id_=None)],
    publisher = Basic(name='DC Comics', id_=None),
    imprint = Basic(name='Vertigo', id_=9),
    cover_date = datetime.date(1994, 12, 1),
    prices = [],
    genres = [],
    comments = 'Just some sample metadata.',
    black_and_white = True,
    story_arcs = [Arc(name='Final Crisis, Inc', id_=None, number=None)],
    characters = [Basic(name='Aquaman', id_=None), Basic(name='Mera', id_=None), Basic(name='Garth', id_=None)],
    teams = [Basic(name='Justice League', id_=None), Basic(name='Infinity, Inc', id_=None)],
    locations = [],
    universes = [Universe(name='ABC', id_=25, designation='Earth 25')],
    credits = [],
    reprints = [],
    tags = [],
    pages = [],\n)"""
    assert str(fake_metadata) == expect_res


def test_metadata_overlay(fake_metadata: Metadata, fake_overlay_metadata: Metadata) -> None:
    md = fake_metadata
    # Fake overlay cover date info.
    md.overlay(fake_overlay_metadata)

    assert md.series.name == "Aquaman"
    assert md.issue == "0"
    assert md.stories == fake_metadata.stories
    assert md.cover_date == date(1994, 10, 1)
    assert md.characters == fake_metadata.characters
    assert md.story_arcs == fake_metadata.story_arcs
    assert md.teams == fake_metadata.teams
    assert md.reprints == fake_overlay_metadata.reprints
    assert md.prices == fake_metadata.prices
    assert md.collection_title == fake_metadata.collection_title
    assert md.universes == fake_metadata.universes
    assert md.imprint.name == fake_metadata.imprint.name
    assert md.imprint.id_ == fake_metadata.imprint.id_


def test_metadata_credits(fake_metadata: Metadata) -> None:
    result = [
        Credit(PETER, [Role(WRITER, primary=True)]),
        Credit(MARTY, [Role(PENCILLER), Role(COVER), Role("Inker")]),
    ]

    md = fake_metadata
    md.add_credit(Credit(PETER, [Role(WRITER, None, True)]))
    md.add_credit(Credit(MARTY, [Role(PENCILLER), Role(COVER)]))
    md.add_credit(Credit(MARTY, [Role("Inker")]))
    # Try to add role for creator a 2nd time
    md.add_credit(Credit(MARTY, [Role(PENCILLER)]))

    assert md.credits == result


good_prices = [
    pytest.param(
        Price(Decimal("2.50"), "CA"),
        Price(Decimal("2.5"), "CA"),
        "Valid 2 letter country code",
    ),
    pytest.param(
        Price(Decimal("1.99"), "Canada"),
        Price(Decimal("1.99"), "CA"),
        "Valid country name",
    ),
    pytest.param(Price(Decimal("3.99")), Price(Decimal("3.99"), "US"), "No country given"),
]


@pytest.mark.parametrize(("price", "expected", "reason"), good_prices)
def test_price_metadata(price: Price, expected: Price, reason: str) -> None:  # noqa: ARG001
    assert price == expected


bad_prices = [
    pytest.param(Decimal("2.5"), "ZZZ", "Invalid country name"),
    pytest.param(Decimal("1"), " ", "Space-only country value"),
    pytest.param(Decimal("5.99"), "ZZ", "Invalid 2 letter country code"),
]


@pytest.mark.parametrize(("amount", "country", "reason"), bad_prices)
def test_invalid_price_metadata(
    amount: Decimal,
    country: str,
    reason: str,  # noqa: ARG001
) -> None:
    with pytest.raises(ValueError):  # noqa: PT011
        Price(amount, country)


bad_gtin = [
    pytest.param(75960620237900411123446, None, "Bad UPC length"),
    pytest.param(None, 97816841565111234, "Bad ISBN"),
]

good_gtin = [
    pytest.param(75960620237900511, None, GTIN(upc=75960620237900511), "Good UPC"),
    pytest.param(None, 9781684156511, GTIN(isbn=9781684156511), "Good ISBN"),
]


@pytest.mark.parametrize(("upc", "isbn", "reason"), bad_gtin)
def test_bad_gtin(
    upc: int | None,
    isbn: int | None,
    reason: str,  # noqa: ARG001
) -> None:
    with pytest.raises(ValueError):  # noqa: PT011
        GTIN(upc, isbn)


@pytest.mark.parametrize(("upc", "isbn", "expected", "reason"), good_gtin)
def test_good_gtin(
    upc: int | None,
    isbn: int | None,
    expected: GTIN,
    reason: str,  # noqa: ARG001
) -> None:
    assert GTIN(upc, isbn) == expected


good_series = [
    pytest.param("Foo Bar", "en", Series("Foo Bar", language="en"), "Good language code"),
    pytest.param("Foo", "German", Series("Foo", language="de"), "Long name search"),
]


@pytest.mark.parametrize(("name", "lang", "expected", "reason"), good_series)
def test_good_series(
    name: str,
    lang: str,
    expected: Series,
    reason: str,  # noqa: ARG001
) -> None:
    assert Series(name, language=lang) == expected


bad_series = [
    pytest.param("Foo", "Fugazi", "Invalid language"),
    pytest.param("Bar", " ", "Space-only language value"),
    pytest.param("Foo", "ZZ", "Invalid 2 letter language code"),
]


@pytest.mark.parametrize(("name", "lang", "reason"), bad_series)
def test_bad_series(name: str, lang: str, reason: str) -> None:  # noqa: ARG001
    with pytest.raises(ValueError):  # noqa: PT011
        Series(name, language=lang)



from __future__ import annotations

import io
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

import rarfile

from darkseid.archivers.archiver import Archiver
from darkseid.exceptions import RarError


class RarArchiver(Archiver):
    """
    Handles archiving operations specific to RAR files.

    This class provides methods for reading, writing, and removing files within a RAR archive.
    """

    def __init__(self: RarArchiver, path: Path) -> None:
        """
        Initializes a RarArchiver object with the provided path.

        Args:
            path (Path): The path associated with the RAR file.

        Returns:
            None
        """

        super().__init__(path)

    def read_file(self: RarArchiver, archive_file: str) -> bytes:
        """
        Reads the contents of a file from the RAR archive.

        Args:
            archive_file (str): The file to read from the archive.

        Returns:
            bytes: The content of the file as bytes.

        Raises:
            RarError: If an error occurs during reading.
        """

        try:
            with rarfile.RarFile(self.path) as rf:
                data: bytes = rf.read(archive_file)
        except rarfile.RarCannotExec as e:
            raise RarError(e) from e
        except io.UnsupportedOperation:
            """If rar directory doesn't contain any data, return None."""
            return b""
        else:
            return data

    def remove_file(self: RarArchiver, archive_file: str) -> bool:  # noqa: ARG002
        """
        Removes a file from the RAR archive.

        Args:
            archive_file (str): The file to remove from the archive.
        Returns:
            bool: False, as RAR files are read-only.
        """

        return False

    def remove_files(self: RarArchiver, filename_lst: list[str]) -> bool:  # noqa: ARG002
        """
        Removes multiple files from the RAR archive.

        Args:
            filename_lst (list[str]): The list of files to remove from the archive.

        Returns:
            bool: False, as RAR files are read-only.
        """

        return False

    def write_file(self: RarArchiver, archive_file: str, data: str) -> bool:  # noqa: ARG002
        """
        Writes data to a file in the RAR archive.

        Args:
            archive_file (str): The file to write to in the archive.
            data (str): The data to write to the file.

        Returns:
            bool: False, as RAR files are read-only.
        """
        return False

    def get_filename_list(self: RarArchiver) -> list[str]:
        """
        Returns a list of filenames in the RAR archive.

        Returns:
            list[str]: A sorted list of filenames in the archive.

        Raises:
            RarError: If an error occurs during retrieval.
        """

        try:
            with rarfile.RarFile(self.path) as rf:
                return sorted(rf.namelist())
        except rarfile.RarCannotExec as e:
            raise RarError(e) from e

    def copy_from_archive(
        self: RarArchiver,
        other_archive: Archiver,  # noqa: ARG002
    ) -> bool:
        """
        Copies files from another archive to the RAR archive.

        Args:
            other_archive (Archiver): The archive to copy files from.

        Returns:
            bool: False, as RAR files are read-only.
        """

        return False



"""Some generic utilities."""
# Copyright 2012-2014 Anthony Beville
# Copyright 2019 Brian Pepple

import itertools
from collections import defaultdict
from pathlib import Path


def get_recursive_filelist(path_list: list[Path]) -> list[Path]:
    """
    Retrieves a list of files recursively from the provided paths.

    Args:
        path_list (list[Path]): List of paths to search for files.

    Returns:
        list[Path]: A sorted list of files found in the provided paths.
    """

    filelist: list[Path] = []
    for path_str in path_list:
        path = Path(path_str)
        if path.is_dir():
            for comic_format in ["*.cbz", "*.cbr"]:
                filelist.extend(iter(path.rglob(comic_format)))
        else:
            filelist.append(path)

    return sorted(filelist)


def list_to_string(list_of_strings: list[str]) -> str:
    """
    Converts a list of strings into a single comma-separated string.

    Args:
        list_of_strings (list[str]): The list of strings to convert.

    Returns:
        str: The comma-separated string.
    """

    return ", ".join((f'"{item}"' if "," in item else item) for item in list_of_strings)


def remove_articles(text: str) -> str:
    """
    Removes common articles from the input text.

    Args:
        text (str): The text from which articles are to be removed.

    Returns:
        str: The text with articles removed.
    """

    articles = frozenset(
        {
            "&",
            "a",
            "am",
            "an",
            "and",
            "as",
            "at",
            "be",
            "but",
            "by",
            "for",
            "if",
            "is",
            "issue",
            "it",
            "it's",
            "its",
            "itself",
            "of",
            "or",
            "so",
            "the",
            "with",
        }
    )
    new_text = "".join(f"{word} " for word in text.split(" ") if word.casefold() not in articles)

    return new_text[:-1]


def unique_file(file_name: Path) -> Path:
    """
    Generates a unique file name by appending a number in parentheses if the original file name already exists.

    Args:
        file_name (Path): The original file name to make unique.

    Returns:
        Path: The unique file name.
    """

    original_stem = file_name.stem
    for i in itertools.count(1):  # noqa: RET503
        if not file_name.exists():
            return file_name
        file_name = file_name.parent / f"{original_stem} ({i}){file_name.suffix}"


def xlate(data: int | str | None, is_int: bool = False) -> int | str | None:
    """
    Translates data to an integer or string based on the provided flag.

    Args:
        data (int | str | None): The data to translate.
        is_int (bool): A flag indicating whether to translate to an integer.

    Returns:
        int | str | None: The translated data.
    """

    if data is None or data == "":
        return None
    if is_int:
        i = str(data).translate(
            defaultdict(
                lambda: None,
                zip((ord(c) for c in "1234567890"), "1234567890", strict=False),
            ),
        )
        if i == "0":
            return "0"
        return int(i) if i else None
    return str(data)



from typing import ClassVar

from darkseid.archivers.archiver import Archiver
from darkseid.archivers.rar import RarArchiver
from darkseid.archivers.zip import ZipArchiver


class UnknownArchiver(Archiver):
    """
    Represents an unknown archiver.

    This class provides functionality for an unknown archiver type.
    """

    @staticmethod
    def name() -> str:
        """
        Returns the name of the archiver as "Unknown".

        Returns:
            str: The name of the archiver.
        """

        return "Unknown"

    __all__: ClassVar[list] = [Archiver, RarArchiver, "UnknownArchiver", ZipArchiver]
